// ===================================================================
//
// CBP_RotMot_PID_1.c
//
// 03-01-21
//
// Basic PID solution with ZN-parameters
//   const double k_p = 24;
//  const double k_i = 24;
//  const double k_d = 6;
//
// Display uses integral time, up to a max of SAMPLING_PERIOD
//
// ===================================================================
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <webots/display.h>
#include <webots/motor.h>
#include <webots/position_sensor.h>
#include <webots/robot.h>
#include <webots/supervisor.h>

#include "CBP_Helper_H.h"

#define TARGET_POSITION 1.500
#define INCREMENT 10
#define SUFFICIENT_NUMBER_OF_PERIODS 10
#define THRESHOLD 0.001
#define SAMPLING_PERIOD 400
#define LABEL_X 0.05
#define LABEL_Y 0.95
#define BLACK 0x000000
#define WHITE 0xFFFFFF
#define RED 0xFF0000
 
// Display variables
static WbDeviceTag display;
static int width = 0;
static int height = 0, half_height = 0;
static double display_middle_y = 0.0;
static double display_unit_x = 0.0;
static double display_unit_y = 0.0;
static int time_step = 0;

// Error related variables
static double error = TARGET_POSITION, previous_error = TARGET_POSITION;
static double max = 0.0, min = 0.0;
double time = 0;

double period, zeroTime1, zeroTime2;

WbDeviceTag motor;
WbDeviceTag position_sensor;
  
static void reset_display() {
  wb_display_set_color(display, BLACK);
  wb_display_fill_rectangle(display, 0, 0, width - 1, height - 1);
  wb_robot_step(time_step);
  wb_display_set_color(display, RED);
  wb_display_draw_line(display, 0, half_height, width, half_height);
  wb_robot_step(time_step);
  wb_display_set_color(display, WHITE);  // draw position error graph in white
}

static void reset() {
  time = 0;
  previous_error = TARGET_POSITION;
  max = 0.0;
  min = 0.0;
  reset_display();
}

static void draw_error(double time) {
  const double previous_t = (time - 1) * display_unit_x;
  const double previous_err = display_middle_y - previous_error * display_unit_y;
  const double current_t = previous_t + display_unit_x;
  const double current_err = display_middle_y - error * display_unit_y;
  wb_display_draw_line(display, (int)previous_t, (int)previous_err, (int)current_t, (int)current_err);
}

void init() {

  // Initializes motor
  motor = wb_robot_get_device("rotational motor");
  wb_motor_set_control_pid(motor, 1.0, 0.0, 0.0);
  // Initializes position sensor
  position_sensor = wb_robot_get_device("position sensor");
  wb_position_sensor_enable(position_sensor, time_step);

  // Initializes display content and display variables
  display = wb_robot_get_device("display");
  height = wb_display_get_height(display);
  half_height = height / 2;
  width = wb_display_get_width(display);
  display_middle_y = 0.5 * height;
  display_unit_x = ((double)width) / SAMPLING_PERIOD;
  display_unit_y = ((double)height) / M_PI;
}


int main(int argc, char **argv) {

  FILE *logfile;
  double Kp, Ki, Kd;
  
  Kp = atof(argv[1]);
  Ki = atof(argv[2]);
  Kd = atof(argv[3]);
  
  printf("Kp = %f, Ki = %f, Kd = %f\n",Kp,Ki,Kd);
  
  wb_robot_init();
  
  logfile = fopen("Plot_PID.m","w");
 // char title[20];
 // sprintf(title,"Kp = %f, Ki = %f, Kd = %f\n",Kp,Ki,Kd);
  writeOctaveHead(logfile);
  
  time_step = wb_robot_get_basic_time_step();
 
  printf("Start: time_step = %d\n",time_step); 
  init();

  //const double k_p = 24;
  //const double k_i = 24;
  //const double k_d = 6;
     
  wb_motor_set_control_pid(motor, Kp, Ki, Kd);
  
  reset();
  
  wb_display_draw_text(display, "PID with tuned gains", width / 3, 3 * height / 4);
  while (wb_robot_step(time_step) != -1) {
    wb_motor_set_position(motor, TARGET_POSITION);
    time += (double)time_step/1000.0;
    double position = wb_position_sensor_get_value(position_sensor);
    error = TARGET_POSITION - position;
    draw_error(1000*time/time_step);
    
    // GEt zeo-crossings and estimate period
    if(previous_error > 0 && error < 0)
      zeroTime1 = time;
    if(previous_error < 0 && error > 0)
      zeroTime2 = time;
    period = 2*(zeroTime2 - zeroTime1);
    printf("Time = %f, Error = %f, Period = %f\n",time,error,period);
    
    previous_error = error;
    writeDataToOctave(logfile,time, position);
    
  }
  writeOctaveTail(logfile,Kp, Ki, Kd);
  fflush(logfile);
  fclose(logfile);
  
  wb_robot_cleanup();
    
  return 0;
}
